
/*****************************************************************************
*
* Copyright (c) 2003-2026 by the esys.escript Group
* https://github.com/LutzGross/esys-escript.github.io
*
* Primary Business: Queensland, Australia
* Licensed under the Apache License, version 2.0
* http://www.apache.org/licenses/LICENSE-2.0
*
* See CREDITS file for contributors and development history
**
*****************************************************************************/

#include "DataReady.h"

namespace escript
{

DataReady::DataReady(const FunctionSpace& what, const ShapeType& shape,
                     bool isDataEmpty)
    : parent(what, shape, isDataEmpty)
{
}


DataReady_ptr DataReady::resolve()
{
    return REFCOUNTNS::dynamic_pointer_cast<DataReady>(this->getPtr());
}



}
