escript_opts_version=203


boost_libs=['boost_python39','boost_numpy39']
boost_prefix=['/usr/include/','/usr/lib/x86_64-linux-gnu/']
build_trilinos='make'
lapack=1
lapack_prefix=['/usr/include/x86_64-linux-gnu','/usr/include/x86_64-linux-gnu/']
mpi4py=1
mpi='OPENMPI'
mpi_prefix='/usr/lib/x86_64-linux-gnu/openmpi/'
mumps=0
mumps_prefix=['/usr/include/x86_64-linux-gnu','/usr/include/x86_64-linux-gnu/']
openmp=1
paso=1
pythoncmd='python3'
pythonlibpath='/usr/lib/x86_64-linux-gnu/'
pythonincpath='/usr/include/python3.9/'
pythonlibname='python3.9'
silo=1
silo_libs=['siloh5','hdf5_serial']
silo_prefix=['/usr/include','/usr/lib/x86_64-linux-gnu/']
umfpack=1
umfpack_prefix=['/usr/include/suitesparse','/usr/lib/x86_64-linux-gnu/']
werror=0
