
/*****************************************************************************
*
* Copyright (c) 2003-2026 by the esys.escript Group
* https://github.com/LutzGross/esys-escript.github.io
*
* Primary Business: Queensland, Australia
* Licensed under the Apache License, version 2.0
* http://www.apache.org/licenses/LICENSE-2.0
*
* See CREDITS file for contributors and development history
**
*****************************************************************************/

#include "Dodgy.h"

namespace escript
{

binOpFnPtr binOpFnPtrFromVoidPtr(void* v)
{
	return reinterpret_cast<binOpFnPtr>(v);

}




void* voidPtrFromBinOpFnPtr(binOpFnPtr f)
{

	return reinterpret_cast<void*>(f);
}





}
